# Copyright (c) 2014, 2015, 2016, 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields.fixedpoint - Fixed-point float to allow float values to be safely used cross-platform and for indexes and filtering.
#    Trying to use a native float will yield different results on different architectures, python versions, etc.
#

# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

from . import IRField, irNull

__all__ = ('IRFixedPointField',)

class IRFixedPointField(IRField):
	'''
		IRFixedPointField - A field which represents a real number (a whole part and a fractional part), such as 3.7 or 3.14159).

	        Use this instead of an IRField(...valueType=float) to get accurate results across platforms, systems, and python versions, and to use for indexing.

		An IRFixedPointField is indexable, and has no option to hash the index.
	'''

	CAN_INDEX = True

	def __init__(self, name='', decimalPlaces=5, defaultValue=irNull):
		'''
			__init__ - Create this object.

			@param name <str> - Field name (or blank if used in an IRFieldChain)

			@param decimalPlaces <int> - The number of decimal places to use (precision). Values will be rounded to this many places, and always have
			  this many digits after the decimal point.

			@param defaultValue - The default value for this field

			An IRFixedPointField is indexable, and has no option to hash the index.
		'''
		self.decimalPlaces = decimalPlaces

		# Make sure the defaultValue gets rounded such that it does not exceed #decimalPlaces
		if isinstance(defaultValue, int):
			defaultValue = float(defaultValue)
		elif isinstance(defaultValue, float):
			defaultValue = round(defaultValue, decimalPlaces)

		self.defaultValue = defaultValue

	def _fromStorage(self, value):
		# Round here in case number of decimalPlaces changes on the field since storage
		return round(float(value), self.decimalPlaces)
	
	def _fromInput(self, value):
		return round(float(value), self.decimalPlaces)

	def _toStorage(self, value):
		if type(value) != float:
			value = float(value)

		return self._getFormatStr() % (value,)

	def _getFormatStr(self):
		return '%.' + str(self.decimalPlaces) + 'f'

	def _getReprProperties(self):
		return [ 'decimalPlaces=%d' %(self.decimalPlaces, ) ]

	def copy(self):
		return self.__class__(name=self.name, decimalPlaces=self.decimalPlaces, defaultValue=self.defaultValue)

	def __new__(self, name='', decimalPlaces=5, defaultValue=irNull):
		return IRField.__new__(self, name)


# vim: set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

import datetime
import sys
import IndexedRedis
from IndexedRedis import IndexedRedisModel, IRField
from IndexedRedis.fields import IRFixedPointField

# vim: ts=4 sw=4 expandtab

class MyFixedPointModel(IndexedRedisModel):

    FIELDS = [ \
        IRField('name'),
        IRFixedPointField('value', decimalPlaces=22),
    ]

    INDEXED_FIELDS = [ \
        'name',
        'value',
    ]

    KEY_NAME = 'MyFixedPointModel'


if __name__ == '__main__':

    MyFixedPointModel.objects.delete()

    pi = 21.991192 / 7.0
    
    newObj = MyFixedPointModel(name='hello', value=pi)

    print ( "My value: <%s>%s\n\n" %(str(type(newObj.value)), repr(newObj.value)))
    print ( "My dict (not for storage): %s\n\n" %(str(newObj.asDict(forStorage=False, strKeys=True)),))
    print ( "My dict (for storage): %s\n\n" %(str(newObj.asDict(forStorage=True, strKeys=True)),))

    newObj.save()

    fetch1 = MyFixedPointModel.objects.filter(name='hello').first()

    if not fetch1:
        print ( "No result for fetch1\n\n" )
    else:
        print ( "fetch1 value: <%s>%s\n\n" %(str(type(fetch1.value)), repr(fetch1.value)))
        print ( "fetch1 dict (not for storage): %s\n\n" %(str(fetch1.asDict(forStorage=False, strKeys=True)),))
        print ( "fetch1 dict (for storage): %s\n\n" %(str(fetch1.asDict(forStorage=True, strKeys=True)),))
        
    fetch2 = MyFixedPointModel.objects.filter(value=pi).first()
    if not fetch2:
        print ( "No result for fetch2\n\n" )
    else:
        print ( "fetch2 value: <%s>%s\n\n" %(str(type(fetch2.value)), repr(fetch2.value)))
        print ( "fetch2 dict (not for storage): %s\n\n" %(str(fetch2.asDict(forStorage=False, strKeys=True)),))
        print ( "fetch2 dict (for storage): %s\n\n" %(str(fetch2.asDict(forStorage=True, strKeys=True)),))



#!/usr/bin/env python

import datetime
import sys
import IndexedRedis
from IndexedRedis import IndexedRedisModel, IRField
from IndexedRedis.fields import IRFieldChain, IRCompressedField, IRBase64Field
from IndexedRedis.fields.FieldValueTypes import IRJsonValue

# vim: ts=4 sw=4 expandtab

class MyChainedModel(IndexedRedisModel):

    FIELDS = [ \
        IRField('name'),
        IRFieldChain('data', [IRField(valueType=IRJsonValue), IRCompressedField(), IRBase64Field()]) ,
    ]

    INDEXED_FIELDS = [ \
        'name',
    ]

    KEY_NAME = 'MyChainedModel'


if __name__ == '__main__':

    MyChainedModel.objects.delete()

    vanillaDict = { 'hello' : 'world', 'deep' : { 'getting' : 'deeper' } }
    
    newObj = MyChainedModel(name='hello', data=vanillaDict)

    print ( "My data: <%s>%s\n\n" %(str(type(newObj.data)), repr(newObj.data)))
    print ( "My dict (not for storage): %s\n\n" %(str(newObj.asDict(forStorage=False, strKeys=True)),))
    print ( "My dict (for storage): %s\n\n" %(str(newObj.asDict(forStorage=True, strKeys=True)),))

    newObj.save()

    fetch1 = MyChainedModel.objects.filter(name='hello').first()

    if not fetch1:
        print ( "No result for fetch1\n\n" )
    else:
        print ( "fetch1 data: <%s>%s\n\n" %(str(type(fetch1.data)), repr(fetch1.data)))
        print ( "fetch1 dict (not for storage): %s\n\n" %(str(fetch1.asDict(forStorage=False, strKeys=True)),))
        print ( "fetch1 dict (for storage): %s\n\n" %(str(fetch1.asDict(forStorage=True, strKeys=True)),))
        
#    fetch2 = MyChainedModel.objects.filter(data=datetime.datetime(1989, 6, 28, 12, 12, 0)).first()
#    if not fetch2:
#        print ( "No result for fetch2\n\n" )
#    else:
#        print ( "fetch2 timestamp: <%s>%s\n\n" %(str(type(fetch2.timestamp)), repr(fetch2.timestamp)))
#        print ( "fetch2 dict (not for storage): %s\n\n" %(str(fetch2.asDict(forStorage=False)),))
#        print ( "fetch2 dict (for storage): %s\n\n" %(str(fetch2.asDict(forStorage=True)),))



#!/usr/bin/env python
#
#  Copyright (c) 2015, 2016, 2017 Tim Savannah under following terms:
#   You may modify and redistribe this script with your project
#
# It will download the latest GoodTests.py and use it to execute the tests.
#
#  This should be placed in a directory, "tests", at the root of your project. It assumes that ../$MY_PACKAGE_MODULE is the path to your test module, and will create a symlink to it in order to run tests.
#  The tests should be found in $MY_TEST_DIRECTORY in given "tests" folder.


# NOTE: Since version 1.2.3, you can also import this (like from a graphical application) and call the "main()" function.
#  All of the following globals are the defaults,  but can be overridden when calling main() (params have the same name as the globals).

import imp
import os

import glob
import subprocess
import sys

# URL to current version of GoodTests.py - You only need to change this if you host an internal copy.
GOODTESTS_URL = 'https://raw.githubusercontent.com/kata198/GoodTests/master/GoodTests.py'

# This should be your module name, and can be any relative or absolute path, or just a module name. 
# If just a module name is given, the directory must be in current directory or parent directory.
MY_PACKAGE_MODULE = 'IndexedRedis'

#  Normally, you want to test the codebase during development, so you don't care about the site-packages installed version.
#     If you want to allow testing with any module by @MY_PACKAGE_MODULE in the python path, change this to True.
ALLOW_SITE_INSTALL = False

# This is the test directory that should contain all your tests. This should be a directory in your "tests" folder
MY_TEST_DIRECTORY = 'UnitTests'

__version__ = '2.1.0'
__version_tuple__ = (2, 1, 0)

def findGoodTests():
    '''
        findGoodTests - Tries to find GoodTests.py

        @return <dict> {
            'path' <str> -> Path to GoodTests.py (for execution)
            'success' <bool> -> True/False if we successfully found GoodTests.py
        }
    '''
    pathSplit = os.environ['PATH'].split(':')
    if '.' not in pathSplit:
        pathSplit = ['.'] + pathSplit
        os.environ['PATH'] = ':'.join(pathSplit)

    result = ''
    success = False
    for path in pathSplit:
        if path.endswith('/'):
            path = path[:-1]
        guess = path + '/GoodTests.py'
        if os.path.exists(guess):
            success = True
            result = guess
            break

    return {
        'path'    : result,
        "success" : success 
    }

def download_goodTests(GOODTESTS_URL=None):
    '''
        download_goodTests - Attempts to download GoodTests, using the default global url (or one provided).
    
        @return <int> - 0 on success (program should continue), otherwise non-zero (program should abort with this exit status)
    '''
    if GOODTESTS_URL is None:
        GOODTESTS_URL = globals()['GOODTESTS_URL']

    validAnswer = False
    while validAnswer == False:
        sys.stdout.write('GoodTests not found. Would you like to install it to local folder? (y/n): ')
        sys.stdout.flush()
        answer = sys.stdin.readline().strip().lower()
        if answer not in ('y', 'n', 'yes', 'no'):
            continue
        validAnswer = True
        answer = answer[0]

    if answer == 'n':
        sys.stderr.write('Cannot run tests without installing GoodTests. http://pypi.python.org/pypi/GoodTests or https://github.com/kata198/Goodtests\n')
        return 1
    try:
        import urllib2 as urllib
    except ImportError:
        try:
            import urllib.request as urllib
        except:
            sys.stderr.write('Failed to import urllib. Trying pip.\n')
            import subprocess
            pipe = subprocess.Popen('pip install GoodTests', shell=True)
            res = pipe.wait()
            if res != 0:
                sys.stderr.write('Failed to install GoodTests with pip ordirect download. aborting.\n')
                return 1
    try:
        response = urllib.urlopen(GOODTESTS_URL)
        contents = response.read()
        if str != bytes:
            contents = contents.decode('ascii')
    except Exception as e:
        sys.stderr.write('Failed to download GoodTests.py from "%s"\n%s\n' %(GOODTESTS_URL, str(e)))
        return 1
    try:
        with open('GoodTests.py', 'w') as f:
            f.write(contents)
    except Exception as e:
        sys.stderr.write('Failed to write to GoodTests.py\n%s\n' %(str(e,)))
        return 1
    try:
        os.chmod('GoodTests.py', 0o775)
    except:
        sys.stderr.write('WARNING: Failed to chmod +x GoodTests.py, may not be able to be executed.\n')

    try:
        import GoodTests
    except ImportError:
        sys.stderr.write('Seemed to download GoodTests okay, but still cannot import. Aborting.\n')
        return 1

    return 0


def main(thisDir=None, additionalArgs=[], MY_PACKAGE_MODULE=None, ALLOW_SITE_INSTALL=None, MY_TEST_DIRECTORY=None, GOODTESTS_URL=None):
    '''
        Do the work - Try to find GoodTests.py, else prompt to download it, then run the tests.

        @param thisDir <None/str> - None to use default (directory this test file is in, or if not obtainable, current directory).
        @param additionalArgs <list> - Any additional args to pass to GoodTests.py

        Remainder of params take their global (top of file) defaults unless explicitly set here. See top of file for documentation.

        @return <int> - Exit code of application. 0 on success, non-zero on failure.

            TODO: Standardize return codes so external applications can derive failure without parsing error strings.
    '''

    if MY_PACKAGE_MODULE is None:
        MY_PACKAGE_MODULE = globals()['MY_PACKAGE_MODULE']
    if ALLOW_SITE_INSTALL is None:
        ALLOW_SITE_INSTALL = globals()['ALLOW_SITE_INSTALL']
    if MY_TEST_DIRECTORY is None:
        MY_TEST_DIRECTORY = globals()['MY_TEST_DIRECTORY']
    if GOODTESTS_URL is None:
        GOODTESTS_URL = globals()['GOODTESTS_URL']
   

    if not thisDir:
        thisDir = os.path.dirname(__file__)

    if not thisDir:
        thisDir = str(os.getcwd())
    elif not thisDir.startswith('/'):
        thisDir = str(os.getcwd()) + '/' + thisDir

    # If GoodTests is in current directory, make sure we find it later
    if os.path.exists('./GoodTests.py'):
        os.environ['PATH'] = str(os.getcwd()) + ':' + os.environ['PATH']

    os.chdir(thisDir)

    didDownload = False
    
    goodTestsInfo = findGoodTests()
    if goodTestsInfo['success'] is False:
        downloadRet = download_goodTests(GOODTESTS_URL)
        if downloadRet != 0:
            return downloadRet
        goodTestsInfo = findGoodTests()
        if goodTestsInfo['success'] is False:
            sys.stderr.write('Could not download or find GoodTests.py. Try to download it yourself using "pip install GoodTests", or wget %s\n' %( GOODTESTS_URL,))
            return 1

    PYTHON_ENDINGS = ('.py', '.pyc', '.pyo')

    origMyPackageModule = MY_PACKAGE_MODULE[:]
    baseName = os.path.basename(MY_PACKAGE_MODULE)
    dirName = os.path.dirname(MY_PACKAGE_MODULE)
    
    newPath = None
    if dirName not in ('.', ''):
        if dirName.startswith('.'):
            dirName = os.getcwd() + os.sep + dirName + os.sep
        newPath = dirName
    elif dirName == '':
        inCurrentDir = False
        try:
            imp.find_module(MY_PACKAGE_MODULE)
            inCurrentDir = True
        except ImportError:
            # COMPAT WITH PREVIOUS runTests.py: Try plain module in parent directory
            foundIt = False
            oldSysPath = sys.path[:]
            sys.path = [os.path.realpath(os.getcwd() + os.sep + '..' + os.sep)]
            try:
                imp.find_module(MY_PACKAGE_MODULE)
                foundIt = True
                sys.path = oldSysPath
            except ImportError as e:
                sys.path = oldSysPath
                if not ALLOW_SITE_INSTALL:
                    sys.stderr.write('Cannot find "%s" locally.\n' %(MY_PACKAGE_MODULE,))
                    return 2
                else:
                    try:
                        __import__(baseName)
                    except:
                        sys.stderr.write('Cannot find "%s" locally or in global python path.\n' %(MY_PACKAGE_MODULE,))
                        return 2

            if foundIt is True:
                newPath = os.path.realpath(os.getcwd() + os.sep + '..' + os.sep)
        if inCurrentDir is True:
            newPath = os.path.realpath(os.getcwd() + os.sep + '..' + os.sep)
    
    if newPath:
        newPythonPath = [newPath] + [x for x in os.environ.get('PYTHONPATH', '').split(':') if x]
        os.environ['PYTHONPATH'] = ':'.join(newPythonPath)
        sys.path = [newPath] + sys.path

    try:
        __import__(baseName)
    except ImportError as e:
        if baseName.endswith(('.py', '.pyc', '.pyo')):
            MY_PACKAGE_MODULE = baseName[baseName.rindex('.')]

        if e.name != MY_PACKAGE_MODULE:
            sys.stderr.write('Error while importing %s: %s\n Likely this is another dependency that needs to be installed\nPerhaps run "pip install %s" or install the providing package.\n\n' %(e.name, str(e), e.name))
            return 1
        sys.stderr.write('Could not import %s. Either install it or otherwise add to PYTHONPATH\n%s\n' %(MY_PACKAGE_MODULE, str(e)))
        return 1

    if not os.path.isdir(MY_TEST_DIRECTORY):
        if not os.path.exists(MY_TEST_DIRECTORY):
            sys.stderr.write('Cannot find test directory: %s\n' %(MY_TEST_DIRECTORY,))
        else:
            sys.stderr.write('Provided test directory, "%s" is not a directory.\n' %(MY_TEST_DIRECTORY,))
        return 3

    sys.stdout.write('Starting test..\n')
    sys.stdout.flush()
    sys.stderr.flush()


    didTerminate = False

    pipe = subprocess.Popen([goodTestsInfo['path']] + additionalArgs + [MY_TEST_DIRECTORY], env=os.environ, shell=False)
    while True:
        try:
            pipe.wait()
            break
        except KeyboardInterrupt:
            if not didTerminate:
                pipe.terminate()
                didTerminate = True
            else:
                pipe.kill()
                break

    return 0


if __name__ == '__main__':
    ret = main(None, sys.argv[1:])
    sys.exit(ret)



# Copyright (c) 2014, 2015, 2016, 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# fields.pickle - Some types and objects related to pickled . Use this in place of IRField ( in FIELDS array ) to activate


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

from . import IRField, irNull
try:
	import cPickle as pickle
except ImportError:
	import pickle

from IndexedRedis.compat_str import isStringy, isEncodedString, isEmptyString

# NOTE: This pickle class originally had implcit base64 encoding and decoding so it could be used for indexes,
#  but even with same protocol python2 and python3, and possibly even different platforms and same version
#  create different pickles for the same objects. Can be as simple as the system supports microseconds,
#  or has additional methods, or whatever, but it's not reliable so don't allow it.

class IRPickleField(IRField):
	'''
		IRPickleField - A field which pickles its data before storage and loads after retrieval.

		This uses the python2 format, so is safe to use between python2 and python3.

		Because even with the same format, python2 and python3 can output different pickle strings for the same object,
		  as well as different host configurations may lead to different output, this field type is not indexable.
	'''

	# Sigh.... so we _can_ index on a pickle'd field, except even with the same protocol the pickling is different between python2 and python3
	CAN_INDEX = False

	def __init__(self, name='', defaultValue=irNull):
		'''
			__init__ - Create an IRPickleField

			@param name <str> - Field name

			@param defaultValue - The default value of this field

			Because even with the same format, python2 and python3 can output different pickle strings for the same object,
			  as well as different host configurations may lead to different output, this field type is not indexable.
		'''
		self.valueType = None
		self.defaultValue = defaultValue

	def _toStorage(self, value):
		if isEmptyString(value):
			return ''

		return pickle.dumps(value, protocol=2)

	def _fromStorage(self, value):
		if isEmptyString(value):
			return ''

		origData = value
		# TODO: Maybe not needed anymore?
		loadedPickle = self.__loadPickle(value)
		if loadedPickle is not None:
			return loadedPickle
		return origData
	
	def _fromInput(self, value):
		return value

	@staticmethod
	def __loadPickle(value):
		if not isEncodedString(value) and isStringy(value):
			return pickle.loads(value)
		return None

	def _getReprProperties(self):
		return []

	def copy(self):
		return self.__class__(name=self.name, defaultValue=self.defaultValue)

	def __new__(self, name='', defaultValue=irNull):
		return IRField.__new__(self, name)

# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRPickle - GoodTests unit tests validating IRPickleField
#

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

# vim: set ts=4 sw=4 expandtab


import sys
import subprocess
from IndexedRedis import IndexedRedisModel
from IndexedRedis.fields import IRPickleField, IRField

# vim: ts=4 sw=4 expandtab

class TestIRField(object):
    '''
        TestIRField - Test some basic IRField stuff
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        # Models should be given different names and key names so they are validated again.
        if testMethod in (self.test_withStrings, self.test_withList):
            class SimpleIRFieldModel(IndexedRedisModel):

                FIELDS = [ IRField('name'), IRPickleField('data') ]
                INDEXED_FIELDS = ['name']

                KEY_NAME = 'Test_SimpleIRPickleFieldModel'

            self.model = SimpleIRFieldModel

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()


    def test_withStrings(self):
        SimpleIRPickleFieldModel = self.model

        myObj = SimpleIRPickleFieldModel(name='test1')

        myObj.save()

        assert not myObj.data , 'Expected undefined data to be False'

        myObj.data = 'hello'

        myObj.save()

        myObjRefetched = SimpleIRPickleFieldModel.objects.filter(name='test1').first()

        assert myObjRefetched.data == 'hello' , 'Expected saved data to be retained'


        myObjRefetched.data = 'goodbye'

        myObjRefetched.save()

        myObj = myObjRefetched

        myObjRefetched = SimpleIRPickleFieldModel.objects.filter(name='test1').first()

        assert myObjRefetched.data == 'goodbye', 'Expected data to be retained'

        myObj2 = SimpleIRPickleFieldModel(name='test2', data='cheese')

        myObj2.save()

        myObj2Refetched = SimpleIRPickleFieldModel.objects.filter(name='test2').first()

        assert myObj2Refetched.data == 'cheese' , 'Expected to be able to specify data in constructor'


        myObj2 = myObj2Refetched

        myObj2.data = ''

        myObj2.save()

        myObj2Refetched = SimpleIRPickleFieldModel.objects.filter(name='test2').first()

        assert not myObj2Refetched.data , 'Expected to be able to remove data on pickled field'


    def test_withList(self):
        '''
            test_withList - Tests using IRPickleField with a list being pickled
        '''
        SimpleIRPickleFieldModel = self.model

        myObj = SimpleIRPickleFieldModel(name='test1')

        myObj.data = ['one', 'two']

        myObj.save()

        myObjRefetched = SimpleIRPickleFieldModel.objects.filter(name='test1').first()

        assert myObjRefetched.data == ['one', 'two'] , 'Expected fetched list data to match what was saved. Got back: %s' %(repr(myObjRefetched.data),)

        myObj = myObjRefetched

        myObj.data.append('three')

        myObj.save()

        assert myObj._origData['data'] == ['one', 'two', 'three'], 'Expected _origData to now contain the latest-saved value, it contains: %s' %(repr(myObj._origData['data']),)

        myObjRefetched = SimpleIRPickleFieldModel.objects.filter(name='test1').first()

        assert myObj.data == ['one', 'two', 'three'] , 'Expected to be able to modify and update a pickled list. Got back: %s' %(repr(myObj.data),)

        bytes_data = b'\x11\x22\x33\x44\xAA\xBB\xCC'
        bytes_data2 = b'\xDD\xEE'

        myObj.data = bytes_data

        myObj.save()

        myObjRefetched = SimpleIRPickleFieldModel.objects.filter(name='test1').first()
        
        assert myObjRefetched.data == bytes_data , 'Expected to be able to save bytes data in IRPickleField'

        myObj = myObjRefetched
        
        myObj.data = bytes_data2
        myObj.save()


        myObjRefetched = SimpleIRPickleFieldModel.objects.filter(name='test1').first()
        
        assert myObjRefetched.data == bytes_data2 , 'Expected to be able to modify bytes data in IRPickleField.\nGot:  \t\t%s\nExpected:\t%s' %(repr(bytes_data2), repr(myObjRefetched.data))

        


if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



# Copyright (c) 2014, 2015, 2016, 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# null - The irNull singleton and IRNullType
#

# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :

import sys

__all__ = ('IR_NULL_STR', 'IR_NULL_BYTES', 'IR_NULL_UNICODE', 'IR_NULL_STRINGS', 'IRNullType', 'irNull')


try:
	unicode
except NameError:
	unicode = str


IR_NULL_STR = 'IRNullType()'
IR_NULL_BYTES = b'IRNullType()'
IR_NULL_UNICODE = u'IRNullType()'

if sys.version_info.major >= 3:
	IR_NULL_STRINGS = (IR_NULL_STR, IR_NULL_BYTES)
else:
	# This generates a unicode warning, but we SHOULDN'T have such a condition.. I don't think
#	IR_NULL_STRINGS = (IR_NULL_STR, IR_NULL_UNICODE)
	IR_NULL_STRINGS = (IR_NULL_STR, )


# TODO: May be indexed as empty string on types that would str the value. empty string != null

# There is an odd "feature" of python 2.7 where the __eq__ method is not called when
#  u'' == irNull
#  however it is in all other forms (including: irNull == u'')
#  
#  when IRNullType extends str. But when it extends unicode, it works as expected.
#
if unicode == str:
	IrNullBaseType = str
else:
	IrNullBaseType = unicode

class IRNullType(IrNullBaseType):
	'''
		The type to represent NULL for anything except string which has no NULL.

		Values of this type only equal other values of this type (i.e. '' does not equal IRNullType())
		Even False does not equal IRNull.

		You probably shouldn't ever need to use this directly, instead use the static instance, "irNull", defined in this module.
	'''

	def __new__(self, val=''):
		'''
			Don't let this be assigned a value.
		'''
		return IrNullBaseType.__new__(self, '')

	def __eq__(self, otherVal):
		return bool(issubclass(otherVal.__class__, IRNullType))
	
	def __ne__(self, otherVal):
		return not bool(issubclass(otherVal.__class__, IRNullType))

	def __str__(self):
		return ''

	def __bool__(self):
		return False
	
	def __nonzero__(self):
		return False
	
	def __repr__(self):
		return 'irNull'


# For all fields which have a type, if they have a null value this will be returned. IRNullType('') != str('') so you can
#  filter out nulls on result like:
#  myObjs = MyModel.objects.all()
#  notNullMyFieldObjs = results.filter(myField__ne=IR_NULL)
global irNull
irNull = IRNullType()

# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :



#!/usr/bin/env python

# Copyright (c) 2014, 2015, 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestModelValidation - GoodTests unit tests validating model validation
#

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

# vim: set ts=4 sw=4 expandtab

import sys
import subprocess
from IndexedRedis import IndexedRedisModel, IRField, InvalidModelException, validatedModels

# vim: ts=4 sw=4 expandtab

# TODO: More tests
class TestModelValidation(object):

    @staticmethod
    def __assertExceptions(name, testMethod, gotException, gotWrongException, contains):
        assert bool(gotWrongException) is False, '[%s] Got exception for "%s", but was not expected (InvalidModelException). Got: " %s "' %(testMethod, name, str(gotWrongException))
        assert bool(gotException) is True, '[%s] Did not get InvalidModelException for "%s", but should have.' %(testMethod, name)
        if contains:
            if type(contains) not in (list, tuple):
                contains = [contains]
            for mustContain in contains:
                assert mustContain  in str(gotException), '[%s] Did not get expected message (containing "%s") for InvalidModelException. Got: " %s "' %(testMethod, mustContain, str(gotException),)

    @staticmethod
    def __assertNoExceptions(name, testMethod, gotException, gotWrongException):
        assert bool(gotException) is False, '[%s] Got InvalidModelException for %s but expected model to be valid. Got: " %s "' %(testMethod, name, str(gotException))
        assert bool(gotWrongException) is False, '[%s] Got an Exception for %s but expected model to be valid. Got: " %s "' %(testMethod, name, str(gotWrongException))

    def __testMethods(self, model, name, contains='', shouldHaveException=True):
        # Create two models so we can validate with two methods
        gotException = False
        gotWrongException = False
        testMethod = 'init'

        validatedModels.clear()
        try:
            myObj = model()
            myObj # Nuke warning
        except InvalidModelException as e:
            gotException = e
        except Exception as e:
            gotWrongException = e

        if shouldHaveException is True:
            self.__assertExceptions(name, testMethod, gotException, gotWrongException, contains)
        else:
            self.__assertNoExceptions(name, testMethod, gotException, gotWrongException)
            

        gotException = False
        gotWrongException = False
        testMethod = 'objects'

        validatedModels.clear()
        try:
            myObj = model.objects.first()
            myObj # Nuke warning
        except InvalidModelException as e:
            gotException = e
        except Exception as e:
            gotWrongException = e


        if shouldHaveException is True:
            self.__assertExceptions(name, testMethod, gotException, gotWrongException, contains)
        else:
            self.__assertNoExceptions(name, testMethod, gotException, gotWrongException)


    def test_NoKeyName(self):
        '''
            test_NoKeyName - Test that lack of key name raises InvalidModelException.
        '''

        class NoKeyModel(IndexedRedisModel):
            FIELDS = [IRField('fielda'), IRField('fieldb')]

            INDEXED_FIELDS = ['fielda']

        self.__testMethods(NoKeyModel, 'no KEY_NAME defined', 'KEY_NAME', shouldHaveException=True)

    def test_NoFields(self):
        '''
            test_NoFields - Test that lack of FIELDS raises InvalidModelException
        '''
        
        class NoFields(IndexedRedisModel):
            KEY_NAME = 'Test_NoFields'

        self.__testMethods(NoFields, 'No Fields', 'FIELDS', shouldHaveException=True)

    def test_InvalidIndexedField(self):
        '''
            test_InvalidIndexedField - Test that a field in INDEXED_FIELD not in FIELDS does not validate.
        '''

        class InvalidIndexedField(IndexedRedisModel):
            FIELDS = [IRField('a'), IRField('b')]
            INDEXED_FIELDS = ['c']
            KEY_NAME = 'Test_InvalidIndexedField'

        self.__testMethods(InvalidIndexedField, 'Invalid Indexed Field', ['INDEXED_FIELDS', 'FIELDS'], shouldHaveException=True)

    def test_ValidModel(self):
        '''
            test_ValidModel - Test that a model passes validation
        '''

        class ValidModel(IndexedRedisModel):
            FIELDS = [IRField('a'), IRField('b')]
            INEXED_FIELDS = ['a']

            KEY_NAME='Test_ValidModel'

        self.__testMethods(ValidModel, 'Valid Model', '', shouldHaveException=False)

    def test_base64FieldsError(self):

        class InvalidModelBase64(IndexedRedisModel):
            FIELDS = [IRField('a'), IRField('b')]
            INDEXED_FIELDS = ['a']

            BASE64_FIELDS = ['b']

            KEY_NAME='Test_InvalidBase64'

        self.__testMethods(InvalidModelBase64, 'Model has removed support for BASE64_FIELDS', '', shouldHaveException=True)


    def test_binaryFieldsError(self):

        class InvalidModelBase64(IndexedRedisModel):
            FIELDS = [IRField('a'), IRField('b')]
            INDEXED_FIELDS = ['a']

            BINARY_FIELDS = ['b']

            KEY_NAME='Test_InvalidBinary'

        self.__testMethods(InvalidModelBase64, 'Model has removed support for BINARY_FIELDS', '', shouldHaveException=True)

            

            
if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



#!/usr/bin/env python

# Copyright (c) 2014, 2015, 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRField - GoodTests unit tests validating model validation
#

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

# vim: set ts=4 sw=4 expandtab

import datetime

import sys
import subprocess

from IndexedRedis import IndexedRedisModel, IRField, irNull, toggleDeprecatedMessages
from IndexedRedis.fields.FieldValueTypes import IRDatetimeValue, IRJsonValue

# vim: ts=4 sw=4 expandtab

class TestIRField(object):
    '''
        TestIRField - Test some basic IRField stuff
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        # Models should be given different names and key names so they are validated again.
        if testMethod in (self.test_simpleIRFieldSaveAndFetch, self.test_updatedFieldsAfterSave):
            class SimpleIRFieldModel(IndexedRedisModel):

                FIELDS = [ IRField('name', valueType=str), IRField('favColour') ]
                INDEXED_FIELDS = ['name', 'favColour']

                KEY_NAME = 'Test_SimpleIRFieldModel'

            self.model = SimpleIRFieldModel
        elif testMethod == self.test_emptyStrNotNull:
            class SimpleIRFieldModel_WithClassicField(IndexedRedisModel):
                FIELDS = [ IRField('name', valueType=str), IRField('favColour'), 'strField' ]
                INDEXED_FIELDS = ['name', 'favColour']

                KEY_NAME = 'Test_SimpleIRFieldModel_WithClassicField'

            self.model = SimpleIRFieldModel_WithClassicField

            toggleDeprecatedMessages(False)

        elif testMethod == self.test_noneFieldValue:
            class SimpleIRFieldModel_NoneField(IndexedRedisModel):
                FIELDS = [ IRField('name'), IRField('nonefield', valueType=None)]
                INDEXED_FIELDS = ['name']

                KEY_NAME = 'Test_SimpleIRFieldModel_NoneField'

            self.model = SimpleIRFieldModel_NoneField
        elif testMethod in (self.test_intFieldValue, self.test_intFieldNulls):
            class SimpleIRFieldModel_Number(IndexedRedisModel):
                FIELDS = [ IRField('name'), IRField('number', valueType=int)]
                INDEXED_FIELDS = ['name', 'number']

                KEY_NAME = 'Test_SimpleIRFieldModel_Number'

            self.model = SimpleIRFieldModel_Number
        elif testMethod == self.test_datetimeFieldValue:
            class SimpleIRFieldModel_Timestamp(IndexedRedisModel):
                FIELDS = [ IRField('name'), IRField('timestamp', valueType=IRDatetimeValue) ]
                INDEXED_FIELDS = ['name', 'timestamp']

                KEY_NAME = 'Test_SimpleIRFieldModel_Timestamp'

            self.model = SimpleIRFieldModel_Timestamp
        elif testMethod == self.test_jsonFieldValue:
            class SimpleIRFieldModel_Json(IndexedRedisModel):
                FIELDS = [ IRField('name'), IRField('jsonData', valueType=IRJsonValue) ]
                INDEXED_FIELDS = ['name']

                KEY_NAME = 'Test_SimpleIRFieldModel_Json'

            self.model = SimpleIRFieldModel_Json

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()

        if testMethod == self.test_emptyStrNotNull:
            toggleDeprecatedMessages(True)

    def test_simpleIRFieldSaveAndFetch(self):
        '''
            test_simpleIRFieldSaveAndFetch - Test that basic IRField's (default type or str type) have same behaviour as a string of field name (pre-IRField)
        '''
        SimpleIRFieldModel = self.model

        myObj = SimpleIRFieldModel(name='Tim', favColour='purple')
        ids = myObj.save()

        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        myObj2 = SimpleIRFieldModel(name='Joe', favColour='blue')

        assert myObj2.hasUnsavedChanges() is True , 'Expected unsaved object to have unsaved changes'

        ids2 = myObj2.save()

        assert myObj2.hasUnsavedChanges() is False , 'Expected saved object to not have unsaved changes'

        assert ids2, 'Expected to get ids returned from save, but did not.'
        assert ids2[0] == myObj2._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleIRFieldModel.objects.get(ids[0])
        assert fetchObj, 'Expected to get an object returned for .get on same id as saved.'
        assert fetchObj._id == ids[0], 'ID on fetched object not same as we requested.'

        assert fetchObj.name == 'Tim', 'Field "name" does not have correct value. Expected "%s", got "%s"' %('Tim', str(fetchObj.name))
        assert fetchObj.favColour == 'purple', 'Field "favColour" does not have correct value'

        fetchObjs2 = SimpleIRFieldModel.objects.filter(favColour='blue').all()

        assert len(fetchObjs2) == 1, 'Expected to get one object back for favColour="blue", got %d' %(len(fetchObjs2),)
        assert (fetchObjs2[0].name == 'Joe' and fetchObjs2[0].favColour == 'blue'), 'Got wrong values back on filter'

        assert fetchObj.delete() == 1, 'Expected delete on fetched object to return 1'

        fetchObj = SimpleIRFieldModel.objects.get(ids[0])
        assert fetchObj is None, 'Expected object to not be present after deleting, but it was.'

    def test_emptyStrNotNull(self):
        '''
            Test that empty str values are not null. Other methods will test that irNull is used when applicable.
        '''
        SimpleIRFieldModel = self.model

        myObj = SimpleIRFieldModel(name='Tim')

        assert myObj.strField != irNull, 'Expected empty classic string field value to not be assigned irNull'
        assert myObj.favColour is irNull, 'Expected empty IRField value to be irNull'

        ids = myObj.save()
        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleIRFieldModel.objects.get(ids[0])
        assert fetchObj, 'Expected to get an object returned for .get on same id as saved.'
        assert fetchObj._id == ids[0], 'ID on fetched object not same as we requested.'

        assert fetchObj.strField != irNull, 'Expected empty classic string field value on fetch to not be assigned irNull'
        assert fetchObj.favColour == irNull, 'Expected IRField not being set to still be irNull'


    def test_noneFieldValue(self):
        '''
            test_noneFieldValue - Use "None" as a type. Expect the field to not be changed.

            This test may be slightly different on python2 vs python3 due to differences in bytes/str type, but should pass on both.
        '''

        SimpleIRFieldModel = self.model

        myObj = SimpleIRFieldModel(name='Tim', nonefield='purple')
        ids = myObj.save()

        assert (type(myObj.nonefield) == str and myObj.nonefield == 'purple'), 'Expected no change for field value on object init, but was changed to %s%s' %(str(type(myObj.nonefield)), repr(myObj.nonefield))

        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        myObj2 = SimpleIRFieldModel(name='Joe', nonefield=b'blue')
        ids2 = myObj2.save()

        assert (type(myObj2.nonefield) == bytes and myObj2.nonefield == b'blue'), 'Expected no change for field value on object init, but was changed to %s%s' %(str(type(myObj2.nonefield)), repr(myObj2.nonefield))

        assert ids2, 'Expected to get ids returned from save, but did not.'
        assert ids2[0] == myObj2._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleIRFieldModel.objects.get(ids[0])

        assert (type(fetchObj.nonefield) == bytes and fetchObj.nonefield == b'purple'), 'Expected field to be returned as bytes (not converted) for nonefield, but was changed to %s%s' %(str(type(fetchObj.nonefield)), repr(fetchObj.nonefield))

        assert fetchObj, 'Expected to get an object returned for .get on same id as saved.'
        assert fetchObj._id == ids[0], 'ID on fetched object not same as we requested.'

        assert fetchObj.name == 'Tim', 'Field "name" does not have correct value. Expected "%s", got "%s"' %('Tim', str(fetchObj.name))
        assert fetchObj.nonefield == b'purple', 'Field "favColour" does not have correct value'


    def test_intFieldValue(self):
        '''
            test_intFieldValue - Test an IRField with "int" as its valueType

        '''
        SimpleIRFieldModel = self.model

        myObj = SimpleIRFieldModel(name='Tim', number=41)
        ids = myObj.save()

        assert (type(myObj.number) == int and myObj.number == 41), 'Expected no change for field value on object init, but was changed to %s%s' %(str(type(myObj.number)), repr(myObj.number))

        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        myObj2 = SimpleIRFieldModel(name='Joe', number=44)
        ids2 = myObj2.save()

        assert ids2, 'Expected to get ids returned from save, but did not.'
        assert ids2[0] == myObj2._id , 'Expected id to be set on saved object, but was not.'

        myObj3 = SimpleIRFieldModel(name='Jim', number=44)
        ids3 = myObj3.save()

        assert ids3, 'Expected to get ids returned from save, but did not.'
        assert ids3[0] == myObj3._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleIRFieldModel.objects.get(ids[0])

        assert type(fetchObj.number) == int, 'Expected field to be returned as bytes (not converted) for number, but was changed to %s%s' %(str(type(fetchObj.number)), repr(fetchObj.number))

        assert fetchObj, 'Expected to get an object returned for .get on same id as saved.'
        assert fetchObj._id == ids[0], 'ID on fetched object not same as we requested.'

        assert fetchObj.name == 'Tim', 'Field "name" does not have correct value. Expected "%s", got "%s"' %('Tim', str(fetchObj.name))
        assert fetchObj.number is 41, 'Field "number" has incorrect value. Expected 41, got: %s' %(repr(fetchObj.number),)

        multipleObjs = SimpleIRFieldModel.objects.filter(number=44).all()

        assert len(multipleObjs) == 2, 'Expected to get two objects for number=44. Got: %d' %(len(multipleObjs),)

        assert ((multipleObjs[0].name == 'Joe' and multipleObjs[1].name == 'Jim') or (multipleObjs[0].name == 'Jim' and multipleObjs[1].name == 'Joe')), 'Expected to get "Joe" and "Jim" objects, but got: %s' %(str([x.name for x in multipleObjs]),)

        SimpleIRFieldModel.objects.filter(number=44).delete()
   
        multipleObjs = SimpleIRFieldModel.objects.filter(number=44).all()

        assert len(multipleObjs) == 0, 'Expected after deleting all with number=44 to get no objects. Got: %d' %(len(multipleObjs),)


        myObj4 = SimpleIRFieldModel(name='Billy')

        assert myObj4.number == irNull, 'Expected unset integer to equal irNull'
        assert myObj4.number != 0, 'Expected unset integer to not equal 0.'
        
        ids4 = myObj4.save()
        assert ids4, 'Expected to get ids returned from save with null value, but did not.'
        assert ids4[0] == myObj4._id , 'Expected id to be set on saved object with null value, but was not.'

        fetchObj4 = SimpleIRFieldModel.objects.get(ids4[0])

        assert fetchObj4._id == myObj4._id , 'Expected fetched object to have same ID as requested.'

        assert fetchObj4.number == irNull, 'Expected unset integer on fetched object to be irNull'
        assert fetchObj4.number != 0, 'Expected unset integer on fetched object to not equal 0'

        fetchObj4.number = 14

        ids4_2 = fetchObj4.save()

        assert ids4_2[0] == ids4[0] , 'Expected update save to not create a new object.'

        fetchObj4_2 = SimpleIRFieldModel.objects.get(ids4[0])
        assert fetchObj4_2.number != irNull , 'Expected once set integer to no longer equal irNull'
        assert fetchObj4_2.number == 14, 'Expected value not saved'
        
    def test_intFieldNulls(self):
        SimpleIRFieldModel = self.model

        myObj = SimpleIRFieldModel(name='Nully')
        ids = myObj.save()

        assert (myObj.number == irNull), 'Expected int field never being set to be irNull'

        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleIRFieldModel(name='Nully')
        assert fetchObj.number == irNull, 'Expected fetched object never have int field set to be irNull'

        myObj.number = 15
        myObj.save()

        assert myObj.number != irNull, 'Expected number not to be null after being set'

        assert myObj.asDict()['number'] != irNull, 'Expected number to not be null after being set'

        fetchObj = SimpleIRFieldModel.objects.filter(name='Nully').all()[0]
        
        assert fetchObj.number == 15, 'Expected number to be set'

        fetchObj.number = irNull
        fetchObj.save()

        fetchObj2 = SimpleIRFieldModel(name='Nully')

        assert fetchObj2.number == irNull, 'Expected after clearing number field it became null'

    def test_datetimeFieldValue(self):
        '''
            test_datetimeFieldValue - Test using an IRDatetimeValue with an IRField.
        '''
        SimpleIRFieldModel = self.model

        myDate = datetime.datetime(1986, 5, 14, 12, 35, 16)

        myObj = SimpleIRFieldModel(name='Tim', timestamp=myDate)

        assert issubclass(myObj.timestamp.__class__, datetime.datetime), 'Expected timestamp to be retained as a datetime.datetime object on init, but was %s%s' %(str(type(myObj.timestamp)), repr(myObj.timestamp))

        ids = myObj.save()

        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        myObj2 = SimpleIRFieldModel(name='Joe', timestamp=datetime.datetime(1988, 2, 13, 14, 15, 16))
        ids2 = myObj2.save()

        assert ids2, 'Expected to get ids returned from save, but did not.'
        assert ids2[0] == myObj2._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleIRFieldModel.objects.get(ids[0])
        assert fetchObj, 'Expected to get an object returned for .get on same id as saved.'
        assert fetchObj._id == ids[0], 'ID on fetched object not same as we requested.'


        assert fetchObj.name == 'Tim', 'Field "name" does not have correct value. Expected "%s", got "%s"' %('Tim', str(fetchObj.name))

        assert issubclass(fetchObj.timestamp.__class__, datetime.datetime), 'Expected timestamp to be a datetime.datetime object after fetch, but was %s%s' %(str(type(fetchObj.timestamp)), repr(fetchObj.timestamp))
        assert fetchObj.timestamp == myDate, 'Expected timestamp retrieved to match original datetime object, but it did not. %s != %s' %(repr(fetchObj.timestamp), repr(myDate))


        fetchedObjs = SimpleIRFieldModel.objects.filter(timestamp=myDate).all()

        assert len(fetchedObjs) == 1, 'Expected to fetch one object for timestamp, got %d' %(len(fetchedObjs),)

        obj = fetchedObjs[0]
        assert obj.name == 'Tim', 'Got wrong fetched object in return.'
        assert obj._id == ids[0] , 'ID is different on fetched object. Expected %d got %s' %(ids[0], str(obj._id))

        assert issubclass(obj.timestamp.__class__, datetime.datetime), 'Expected timestamp to be a datetime.datetime object after fetch, but was %s%s' %(str(type(obj.timestamp)), repr(obj.timestamp))
        assert obj.timestamp == myDate, 'Expected timestamp retrieved to match original datetime object, but it did not. %s != %s' %(repr(obj.timestamp), repr(myDate))

        newDate = datetime.datetime(1987, 3, 13, 1, 1, 1)

        obj.timestamp = newDate
        obj.save()

        fetchedObjs = SimpleIRFieldModel.objects.filter(timestamp=myDate).all()

        assert len(fetchedObjs) == 0, 'Expected to fetch no objects filtering on old date, after date was updated. Got %d: %s' %(len(fetchedObjs), repr([x.asDict() for x in fetchedObjs]))

        fetchedObjs = SimpleIRFieldModel.objects.filter(timestamp=newDate).all()

        assert len(fetchedObjs) == 1, 'Expected to fetch 1 object with new date after updating. Got %s: %s' %(len(fetchedObjs), repr([x.asDict() for x in fetchedObjs]))
        assert fetchedObjs[0].name == 'Tim' and fetchedObjs[0].timestamp == newDate , 'Got wrong values from fetched object after update.'

        noTimestampObj = SimpleIRFieldModel(name='Billy')

        assert noTimestampObj.timestamp == irNull, 'Expected object created with empty timestamp field to have timestamp be assigned irNull'

        ids = noTimestampObj.save()

        assert ids and ids[0] == noTimestampObj._id , 'Did not set ids properly on saved object with a null field'

        fetchedObjs = SimpleIRFieldModel.objects.filter(name='Billy').all()

        assert len(fetchedObjs) == 1, 'Expected to only get one object with name="Billy", but got %d' %(len(fetchedObjs),)

        fetchedObj = fetchedObjs[0]

        assert fetchedObj.timestamp == irNull, 'Expected fetched object with empty timestamp field to have timestamp be assigned irNull'

        fetchedObjs = SimpleIRFieldModel.objects.filter(timestamp=irNull).all()

        assert len(fetchedObjs) == 1 and fetchedObjs[0]._id == ids[0], 'Expected to be able to fetch using irNull as a filter value'


    def test_jsonFieldValue(self):
        SimpleIRFieldModel = self.model

        myData = { "one" : "two",
            "sub" : {
                "level2" : {
                    "val" : 5,
                    "cheese" : "cheddar",
                }
            }
        }

        myObj = SimpleIRFieldModel(name='Tim', jsonData=myData)

        assert issubclass(myObj.jsonData.__class__, dict), 'Expected jsonData to be retained as a dict object on init, but was %s%s' %(str(type(myObj.jsonData)), repr(myObj.jsonData))

        ids = myObj.save()

        assert ids, 'Expected to get ids returned from save, but did not.'
        assert ids[0] == myObj._id , 'Expected id to be set on saved object, but was not.'

        myObj2 = SimpleIRFieldModel(name='Joe', jsonData={'basic' : 'yes'})
        ids2 = myObj2.save()

        assert ids2, 'Expected to get ids returned from save, but did not.'
        assert ids2[0] == myObj2._id , 'Expected id to be set on saved object, but was not.'

        fetchObj = SimpleIRFieldModel.objects.get(ids[0])
        assert fetchObj, 'Expected to get an object returned for .get on same id as saved.'
        assert fetchObj._id == ids[0], 'ID on fetched object not same as we requested.'

        assert fetchObj.name == 'Tim', 'Field "name" does not have correct value. Expected "%s", got "%s"' %('Tim', str(fetchObj.name))

        assert issubclass(fetchObj.jsonData.__class__, dict), 'Expected jsonData to be a dict object after fetch, but was %s%s' %(str(type(fetchObj.jsonData)), repr(fetchObj.jsonData))
        
        fetchedData = fetchObj.jsonData

        assert fetchedData.get('one', '') == 'two', 'jsonData has corrupted values. "one" field should have equaled "two", but equaled: %s' %(fetchedData.get('one', 'None'),)

        try:
            assert fetchedData['sub']['level2']['val'] == 5, 'Got wrong value, expected "val" to retain integer 5, but got: %s%s' %(str(type(fetchedData['sub']['level2']['val'])), repr(fetchedData['sub']['level2']['val']))
        except KeyError as e:
            raise AssertionError('Got key error accessing sub->level2->val: %s' %(str(e),))

        emptyJsonObj = SimpleIRFieldModel(name='Billy')

        assert emptyJsonObj.jsonData == irNull, 'Expected object with no value assigned for jsonData to be irNull'
        
        emptyJsonObj.save()

        emptyFetchObj = SimpleIRFieldModel.objects.get(emptyJsonObj._id)

        assert emptyFetchObj.name == 'Billy' , 'Fetched wrong object'
        assert emptyFetchObj.jsonData == irNull, 'Expected fetched json object with no vlaue assigned to be irNull'

        blankJsonObj = SimpleIRFieldModel(name='Billy', jsonData={})
       
        assert blankJsonObj.jsonData != irNull, 'Expected empty dict to not be converted to irNull'

        blankJsonObj.jsonData['cheese'] = 'swiss'
        blankJsonObj.save()

        fetchObj = SimpleIRFieldModel.objects.get(blankJsonObj._id)

        assert fetchObj.jsonData.get('cheese', None) == 'swiss' , 'Expected updated field to retain new value'

    def test_DatetimeMicrosecond(self):
        datetimeStr = '2015-05-12 12:00:00'
        datetimeObj = datetime.datetime(year=2015, month=5, day=12, hour=12, minute=0, second=0)

        val = IRDatetimeValue(datetimeStr)
        assert val == datetimeObj, 'Expected datetime object created from string to match equivalent'

        datetimeStr += '.203511'
        try:
            val = IRDatetimeValue(datetimeStr)
        except Exception as e:
            raise AssertionError('Expected to be able to create object including microseconds. Got %s %s' %(str(e.__class__.__name__), str(e)))

        val = IRDatetimeValue(datetimeStr)
        assert val == datetimeObj, 'Expected datetime object created from string including microseconds to match equivalent without microseconds'

    def test_updatedFieldsAfterSave(self):
    
        SimpleIRFieldModel = self.model

        obj = SimpleIRFieldModel()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {}, 'Expected empty updatedFields on newly created object'

        obj.favColour = 'purple'

        updatedFields = obj.getUpdatedFields()
        assert 'favColour' in updatedFields , 'Expected favColour to show up in updatedFields after updating value'

        assert not updatedFields['favColour'][0]  , 'Expected old value in updatedFields["favColour"] to be False'
        assert updatedFields['favColour'][1] == 'purple' , 'Expected new value to be in updatedFields["favColour"][1]'

        obj.name = 'hello'

        obj.save()

        updatedFields = obj.getUpdatedFields()
        assert updatedFields == {}, 'Expected empty updatedFields after saving'

        fetchedObj = SimpleIRFieldModel.objects.filter(name='hello').first()

        assert fetchedObj , 'Expected to be able to fetch'

        obj = fetchedObj

        updatedFields = obj.getUpdatedFields()
        assert updatedFields == {}, 'Expected empty updatedFields after fetching'

        obj.favColour = 'purple'
        updatedFields = obj.getUpdatedFields()
        assert updatedFields == {}, 'Expected empty updatedFields after changing to same value'
        

        obj.favColour = 'red'
        updatedFields = obj.getUpdatedFields()

        assert updatedFields['favColour'][0] == 'purple' , 'Expected old value to be in updatedFields["favColour"][0]'
        assert updatedFields['favColour'][1] == 'red' , 'Expected new value to be in updatedFields["favColour"][1]'

        obj.save()
        updatedFields = obj.getUpdatedFields()
        assert updatedFields == {}, 'Expected empty updatedFields after updating'



if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRDefaultValues - Test default values for fields
#

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

import sys
import subprocess

from IndexedRedis import IndexedRedisModel, IRField, irNull
from IndexedRedis.fields import IRClassicField

# vim: ts=4 sw=4 expandtab

class TestIRDefaultValues(object):
    '''
        TestIRField - Test some basic IRField stuff
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        if testMethod == self.test_basicDefaults:
            class Model_Basic_Defaults(IndexedRedisModel):
                
                FIELDS = [
                    IRField('name'),
                    IRField('someNum', valueType=int, defaultValue=-1),
                    IRField('someStr', valueType=str, defaultValue='unset'),
                    IRField('gblDefault'),
                    IRClassicField('classic')
                ]

                INDEXED_FIELDS = ['name', 'someNum', 'someStr', 'gblDefault', 'classic']

                KEY_NAME='TestIRDefaultValues__basicDefaults_1'

            self.model = Model_Basic_Defaults

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()


    @staticmethod
    def _assert_fieldEquals(obj, fieldName, expectedValue, whatExpectedStr='', expectedValueReadable=None):
        val = getattr(obj, fieldName)

        if expectedValueReadable is None:
            expectedValueReadable = repr(expectedValue)

        assert val == expectedValue , 'Expected %s to be %s. Got <%s> %s' %(whatExpectedStr or fieldName, expectedValueReadable, val.__class__.__name__, repr(val))

    def test_basicDefaults(self):
        '''
            test_basicDefaults - Test basic defaultValue functionality
        '''
        
        Model = self.model

        obj = Model()

        self._assert_fieldEquals(obj, 'name', irNull, 'default for IRField without explicit defaultValue', 'irNull')
        self._assert_fieldEquals(obj, 'someNum', -1,  'default for IRField(valueType=int, defaultValue=-1)')
        self._assert_fieldEquals(obj, 'someStr', 'unset',  'default for IRField(valueType=str, defaultValue="unset")')
        self._assert_fieldEquals(obj, 'gblDefault', irNull, 'default for IRField without explicit defaultValue', 'irNull')
        self._assert_fieldEquals(obj, 'classic', "", 'default for IRClassicField', '"" (empty string)')

        assert not obj.getUpdatedFields() , 'Expected no fields to be listed as "updated" when none have been set. Got: %s' %(repr(obj.getUpdatedFields()), )

        obj.name = 'fetch1'
        obj.save()

        self._assert_fieldEquals(obj, 'name', 'fetch1', 'after saving value for IRField')
        self._assert_fieldEquals(obj, 'someNum', -1,  'after saving for IRField(valueType=int, defaultValue=-1)')
        self._assert_fieldEquals(obj, 'someStr', 'unset',  'after saving for IRField(valueType=str, defaultValue="unset")')
        self._assert_fieldEquals(obj, 'gblDefault', irNull, 'after saving for IRField without explicit defaultValue', 'irNull')
        self._assert_fieldEquals(obj, 'classic', "", 'after saving for IRClassicField', '"" (empty string)')


        objFetched = Model.objects.filter(name='fetch1').first()

        assert objFetched , 'Expected to be able to fetch object from name="fetch1"'

        obj = objFetched


        self._assert_fieldEquals(obj, 'name', 'fetch1', 'after fetching value for IRField')
        self._assert_fieldEquals(obj, 'someNum', -1,  'after fetching for IRField(valueType=int, defaultValue=-1)')
        self._assert_fieldEquals(obj, 'someStr', 'unset',  'after fetching for IRField(valueType=str, defaultValue="unset")')
        self._assert_fieldEquals(obj, 'gblDefault', irNull, 'after fetching for IRField without explicit defaultValue', 'irNull')
        self._assert_fieldEquals(obj, 'classic', "", 'after fetching for IRClassicField', '"" (empty string)')

        obj.someNum = 22
        obj.someStr = 'hello goodbye'
        obj.gblDefault = 'zz'
        obj.classic = irNull

        obj.save()

        self._assert_fieldEquals(obj, 'name', 'fetch1', 'after updating value for IRField')
        self._assert_fieldEquals(obj, 'someNum', 22,  'after changing IRField(valueType=int, defaultValue=-1) to non-default value')
        self._assert_fieldEquals(obj, 'someStr', 'hello goodbye',  'after changing  IRField(valueType=str, defaultValue="unset") to non-default value')
        self._assert_fieldEquals(obj, 'gblDefault', 'zz', 'after changing IRField without explicit defaultValue to non-default value')
        self._assert_fieldEquals(obj, 'classic', irNull, 'after changing IRClassicField from default to irNull', 'irNull')

        objFetched = Model.objects.filter(name='fetch1').first()

        assert objFetched , 'Expected to be able to fetch object from name="fetch1"'

        obj = objFetched


        self._assert_fieldEquals(obj, 'name', 'fetch1', 'fetched after updating value for IRField')
        self._assert_fieldEquals(obj, 'someNum', 22,  'fetched after changing IRField(valueType=int, defaultValue=-1) to non-default value')
        self._assert_fieldEquals(obj, 'someStr', 'hello goodbye',  'fetched after changing  IRField(valueType=str, defaultValue="unset") to non-default value')
        self._assert_fieldEquals(obj, 'gblDefault', 'zz', 'fetched after changing IRField without explicit defaultValue to non-default value')
        self._assert_fieldEquals(obj, 'classic', irNull, 'fetched after changing IRClassicField from default to irNull', 'irNull')


if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



#!/usr/bin/env python

# Copyright (c) 2017 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#
# TestIRBase64Field - Test default values for fields
#

# vim: set ts=4 sw=4 st=4 expandtab

# Import and apply the properties (like Redis connection parameters) for this test.
import TestProperties

import base64
import sys
import subprocess

from IndexedRedis import IndexedRedisModel, irNull
from IndexedRedis.compat_str import tobytes
from IndexedRedis.fields import IRBase64Field, IRField

# vim: ts=4 sw=4 expandtab

class TestIRBase64Field(object):
    '''
        TestIRBase64Field - Test base64 field
    '''

    KEEP_DATA = False

    def setup_method(self, testMethod):
        '''
            setup_method - Called before every method. Should set "self.model" to the model needed for the test.
  
            @param testMethod - Instance method of test about to be called.
        '''
        self.model = None

        if testMethod == self.test_general:
            class Model_Base64Value(IndexedRedisModel):
                
                FIELDS = [
                    IRField('name'),
                    IRBase64Field('value', defaultValue=irNull),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME='TestIRBase64Field__ModelBase64Value'

            self.model = Model_Base64Value
        elif testMethod == self.test_defaultValue:
            class Model_Base64DefaultValue(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRBase64Field('value', defaultValue=b'woobley'),
                ]

                INDEXED_FIELDS = ['name']

                KEY_NAME = 'TestIRBase64Field__ModelBase64DefaultValue'

            self.model = Model_Base64DefaultValue

        elif testMethod == self.test_index:
            class Model_Base64Index(IndexedRedisModel):
                FIELDS = [
                    IRField('name'),
                    IRBase64Field('value'),
                ]

                INDEXED_FIELDS = ['name', 'value']

                KEY_NAME = 'TestIRBase64Field__ModelBase64Index'

            self.model = Model_Base64Index

        # If KEEP_DATA is False (debug flag), then delete all objects before so prior test doesn't interfere
        if self.KEEP_DATA is False and self.model:
            self.model.deleter.destroyModel()

    def teardown_method(self, testMethod):
        '''
            teardown_method - Called after every method.

                If self.model is set, will delete all objects relating to that model. To retain objects for debugging, set TestIRField.KEEP_DATA to True.
        '''

        if self.model and self.KEEP_DATA is False:
            self.model.deleter.destroyModel()


    def test_general(self):
        
        Model = self.model

        obj = Model()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected no updated fields when object is first created.\nExpected: %s\nGot:     %s' %(repr([]), repr(updatedFields) )

        assert obj.value == irNull , 'Expected default value of IRBase64Field to be irNull when defaultValue=irNull'

        obj.name = 'one'

        obj.save()

        assert obj.getUpdatedFields() == {} , 'Expected no updated fields after object is saved'

        obj.value = 'Hello World'

        assert obj.value == b'Hello World' , 'Expected IRBase64Field value to be "Hello World" after setting to "Hello World"'

        b64Value = base64.b64encode(tobytes(obj.value))

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == b'Hello World', 'Expected asDict(forStorage=False) to contain IRBase64Field value as original string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == b64Value , 'Expected asDict(forStorage=True) to contain IRBase64Field that was base64 encoded.\nExpected: %s\nGot:     %s' %(repr(b64Value), repr(dictForStorage['value']) )

        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating'

        assert updatedFields['value'][0] == irNull , 'Expected old value to be irNull in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == b'Hello World' , 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving.'

        fetchObj = Model.objects.filter(name='one').first()


        assert fetchObj , 'Expected to be able to fetch object on name="one" after saving.'

        obj = fetchObj

        assert obj.value == b'Hello World' , 'Expected value of fetched to be converted string (As bytes), b"Hello World". Got: %s' %(repr(fetchObj.value), )

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after fetching'

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        assert dictConverted['value'] == b'Hello World', 'After fetching, Expected asDict(forStorage=False) to contain IRBase64Field value as original string (as bytes). Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == b64Value , 'After fetching, Expected asDict(forStorage=True) to contain IRBase64Field that was base64 encoded.\nExpected: %s\nGot:     %s' %(repr(b64Value), repr(dictForStorage['value']) )

        obj.value = b'q123'


        updatedFields = obj.getUpdatedFields()

        assert 'value' in updatedFields , 'Expected "value" to show in updated fields after updating on fetched object'

        assert updatedFields['value'][0] == b"Hello World" , 'Expected old value to be b"Hello World" in updatedFields. Got: %s' %(repr(updatedFields['value'][0]), )
        assert updatedFields['value'][1] == b'q123' , 'Expected converted value to be new value in updatedFields. Got: %s' %(repr(updatedFields['value'][1]), )

        try:
            dictConverted = obj.asDict(forStorage=False, strKeys=True)
            dictForStorage = obj.asDict(forStorage=True, strKeys=True)
        except Exception as e:
            raise AssertionError('Expected to be able to convert to dict for both storage and non-storage. Got exception: %s %s' %(e.__class__.__name__, str(e)))

        b64Value = base64.b64encode(tobytes(obj.value))

        assert dictConverted['value'] == b'q123', 'After fetching, then updating, Expected asDict(forStorage=False) to contain IRBase64Field value as bytes string. Got: %s' %(dictConverted['value'], )
        assert dictForStorage['value'] == b64Value , 'After fetching, then updating, Expected asDict(forStorage=True) to contain IRBase64Field that was base64 encoded.\nExpected: %s\nGot:     %s' %(repr(b64Value), repr(dictForStorage['value']) )

        
        obj.save()

        updatedFields = obj.getUpdatedFields()

        assert updatedFields == {} , 'Expected updatedFields to be clear after saving'

        obj = Model(value='popo')

        assert obj.value == b'popo' , 'Expected constructor to work to set values on IRBase64Field'

    def test_defaultValue(self):

        Model = self.model

        obj = Model()

        assert obj.value == b'woobley' , 'Expected defaultValue to be applied to a base64 field.\nExpected: b"woobley"\nGot:     %s' %(repr(obj.value), )

        obj.name = 'test'

        obj.save()

        assert obj.value == b'woobley' , 'Expected defaultValue to remain on a base64 field after saving'

        objFetched = Model.objects.filter(name='test').first()
        assert objFetched , 'Expected to be able to fetch object'

        obj = objFetched

        assert obj.value == b'woobley' , 'Expected defaultValue to remain on a base64 field after fetching'

        obj.value = b'cheesy'

        obj.save()

        objFetched = Model.objects.filter(name='test').first()
        assert objFetched , 'Expected to be able to fetch object'

        obj = objFetched

        assert obj.value == b'cheesy' , 'Expected to be able to change value from default.'

    def test_index(self):

        Model = self.model

        obj = Model()

        otherObj = Model(name='two', value='val2')
        
        ids = otherObj.save()

        assert ids and ids[0] , 'Failed to save object'

        obj.name = 'one'

        ids = obj.save()
        assert ids and ids[0] , 'Failed to save object'


        fetchedObjs = Model.objects.filter(value=irNull).all()

        assert len(fetchedObjs) == 1 , 'Expected to be able to fetch an IRBase64Field with irNull as the index'
        assert fetchedObjs[0].name == 'one' , 'Fetched wrong object'

        assert fetchedObjs[0].value == irNull , 'Expected after fetch for irNull to be retained.'

        emptyStrObj = Model(value='')

        assert emptyStrObj.value == b'' , 'Expected IRBase64Field to translate empty string to b"" '

        emptyStrObj.name = 'emptystr'
        ids = emptyStrObj.save()

        assert ids and ids[0] , 'Failed to save object'

        fetchedObjs = Model.objects.filter(value=irNull).all()

        assert len(fetchedObjs) == 1 , 'Expected for index on IRBase64Field to treat irNull and empty string as different. fetched both on irNull'


        fetchedObjs = Model.objects.filter(value='').all()

        assert len(fetchedObjs) == 1 , 'Expected for index on IRBase64Field to treat irNull and empty string as different. fetched both on empty str'
        

        obj.value = 'val1'
        ids = obj.save()
        assert ids and ids[0] , 'Failed to save object'

        fetchedObjs = Model.objects.filter(value='val2').all()

        assert len(fetchedObjs) == 1 , 'Expected to be able to fetch object using IRBase64Field with a value'
        assert fetchedObjs[0].name == 'two' , 'Fetched wrong object'


if __name__ == '__main__':
    sys.exit(subprocess.Popen('GoodTests.py -n1 "%s" %s' %(sys.argv[0], ' '.join(['"%s"' %(arg.replace('"', '\\"'), ) for arg in sys.argv[1:]]) ), shell=True).wait())

# vim: set ts=4 sw=4 expandtab



# Copyright (c) 2016 Timothy Savannah under LGPL version 2.1. See LICENSE for more information.
#  
# Some random utility functions


# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :


def hashDictOneLevel(myDict):
	'''
		A function which can generate a hash of a one-level 
		  dict containing strings (like REDIS_CONNECTION_PARAMS)

		@param myDict <dict> - Dict with string keys and values

		@return <long> - Hash of myDict
	'''
	keys = [str(x) for x in myDict.keys()]
	keys.sort()

	lst = []
	for key in keys:
		lst.append(str(myDict[key]) + '__~~__')

	return '+_[,'.join(lst).__hash__()



def _hashList(lst):
	ret = []

	for i in range(len(lst)):

		item = lst[i]

		if issubclass(item.__class__, dict):
			item = [ (k, v) for k, v in item.items() ]

		if issubclass(item.__class__, (list, tuple, set)):
			ret += _hashList(item)
		else:
			try:
				thisHash = hash(item)
			except TypeError as e:
				# Hmm.. 
				import sys
				sys.stderr.write('Got TypeError: %s\n\n' %(str(e), ))
				thisHash = hash( '<%s> %s' %( item.__class__.__name__, repr(item) ) )

			ret.append( str(thisHash) )
	
	return ret



def hashList(lst):
	hashStrs = [ str(hash('[]')) ] + _hashList(lst)

	return hash( '____FS____'.join(hashStrs) )



class KeyList(list):
	'''
		KeyList - A list which is indexable by both values and integer indexes.
	'''
	
	def __getitem__(self, item):
		if isinstance(item, (int, slice)):
			return list.__getitem__(self, item)
		try:
			idx = self.index(item)
		except:
			raise KeyError('No such key in list: %s' %(repr(item), ))

		return list.__getitem__(self, idx)

# vim:set ts=8 shiftwidth=8 softtabstop=8 noexpandtab :
